//
// Copyright © 2022 Hardcore Engineering Inc.
//
// Licensed under the Eclipse Public License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License. You may
// obtain a copy of the License at https://www.eclipse.org/legal/epl-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//
// See the License for the specific language governing permissions and
// limitations under the License.
//

import { AccountRole, DOMAIN_TX, makeCollaborativeDoc, TxOperations, type Ref, type Status } from '@hcengineering/core'
import { leadId, type Lead } from '@hcengineering/lead'
import {
  tryMigrate,
  tryUpgrade,
  type MigrateOperation,
  type MigrationClient,
  type MigrationUpgradeClient,
  type ModelLogger
} from '@hcengineering/model'
import core, { DOMAIN_SPACE } from '@hcengineering/model-core'

import contact, { DOMAIN_CONTACT } from '@hcengineering/model-contact'
import task, { DOMAIN_TASK, createSequence, migrateDefaultStatusesBase } from '@hcengineering/model-task'

import lead from './plugin'
import { defaultLeadStatuses } from './spaceType'

async function createSpace (tx: TxOperations): Promise<void> {
  const current = await tx.findOne(core.class.Space, {
    _id: lead.space.DefaultFunnel
  })
  if (current === undefined) {
    await tx.createDoc(
      lead.class.Funnel,
      core.space.Space,
      {
        name: 'Funnel',
        description: 'Default funnel',
        private: false,
        archived: false,
        members: [],
        type: lead.template.DefaultFunnel
      },
      lead.space.DefaultFunnel
    )
  }
}

async function createDefaults (tx: TxOperations): Promise<void> {
  await createSpace(tx)
  await createSequence(tx, lead.class.Lead)
}

async function migrateIdentifiers (client: MigrationClient): Promise<void> {
  const docs = await client.find<Lead>(DOMAIN_TASK, { _class: lead.class.Lead, identifier: { $exists: false } })
  for (const doc of docs) {
    await client.update(
      DOMAIN_TASK,
      { _id: doc._id },
      {
        identifier: `LEAD-${doc.number}`
      }
    )
  }
}

async function migrateDefaultStatuses (client: MigrationClient, logger: ModelLogger): Promise<void> {
  const defaultTypeId = lead.template.DefaultFunnel
  const typeDescriptor = lead.descriptors.FunnelType
  const baseClass = lead.class.Funnel
  const defaultTaskTypeId = lead.taskType.Lead
  const taskTypeClass = task.class.TaskType
  const baseTaskClass = lead.class.Lead
  const statusAttributeOf = lead.attribute.State
  const statusClass = core.class.Status
  const getDefaultStatus = (oldStatus: Status): Ref<Status> | undefined => {
    return defaultLeadStatuses.find(
      (defStatus) =>
        defStatus.category === oldStatus.category &&
        (defStatus.name.toLowerCase() === oldStatus.name.trim().toLowerCase() ||
          (defStatus.name === 'Negotiation' && oldStatus.name === 'Negotation'))
    )?.id
  }

  await migrateDefaultStatusesBase<Lead>(
    client,
    logger,
    defaultTypeId,
    typeDescriptor,
    baseClass,
    defaultTaskTypeId,
    taskTypeClass,
    baseTaskClass,
    statusAttributeOf,
    statusClass,
    getDefaultStatus
  )
}

async function migrateDefaultTypeMixins (client: MigrationClient): Promise<void> {
  const oldSpaceTypeMixin = `${lead.template.DefaultFunnel}:type:mixin`
  const newSpaceTypeMixin = lead.mixin.DefaultFunnelTypeData
  const oldTaskTypeMixin = `${lead.taskType.Lead}:type:mixin`
  const newTaskTypeMixin = lead.mixin.LeadTypeData

  await client.update(
    DOMAIN_TX,
    {
      objectClass: core.class.Attribute,
      'attributes.attributeOf': oldSpaceTypeMixin
    },
    {
      $set: {
        'attributes.attributeOf': newSpaceTypeMixin
      }
    }
  )

  await client.update(
    DOMAIN_SPACE,
    {
      _class: lead.class.Funnel,
      [oldSpaceTypeMixin]: { $exists: true }
    },
    {
      $rename: {
        [oldSpaceTypeMixin]: newSpaceTypeMixin
      }
    }
  )

  await client.update(
    DOMAIN_TASK,
    {
      _class: lead.class.Lead,
      [oldTaskTypeMixin]: { $exists: true }
    },
    {
      $rename: {
        [oldTaskTypeMixin]: newTaskTypeMixin
      }
    }
  )
}

async function migrateDefaultProjectOwners (client: MigrationClient): Promise<void> {
  const workspaceOwners = await client.model.findAll(contact.class.PersonAccount, {
    role: AccountRole.Owner
  })

  await client.update(
    DOMAIN_SPACE,
    {
      _id: lead.space.DefaultFunnel
    },
    {
      $set: {
        owners: workspaceOwners.map((it) => it._id)
      }
    }
  )
}

export const leadOperation: MigrateOperation = {
  async preMigrate (client: MigrationClient, logger: ModelLogger): Promise<void> {
    await tryMigrate(client, leadId, [
      {
        state: 'migrate-default-statuses',
        func: (client) => migrateDefaultStatuses(client, logger)
      }
    ])
  },
  async migrate (client: MigrationClient): Promise<void> {
    await tryMigrate(client, leadId, [
      {
        state: 'identifier',
        func: migrateIdentifiers
      },
      {
        state: 'migrate-default-type-mixins',
        func: async (client) => {
          await migrateDefaultTypeMixins(client)
        }
      },
      {
        state: 'migrateDefaultProjectOwners',
        func: migrateDefaultProjectOwners
      },
      {
        state: 'migrate-customer-description',
        func: async (client) => {
          await client.update(
            DOMAIN_CONTACT,
            {
              [lead.mixin.Customer + '.description']: { $exists: true }
            },
            {
              $rename: {
                [lead.mixin.Customer + '.description']: lead.mixin.Customer + '.customerDescription'
              }
            }
          )
          const it = await client.traverse(DOMAIN_CONTACT, {
            _class: contact.class.Organization,
            description: { $exists: false }
          })
          while (true) {
            const docs = await it.next(50)
            if (docs == null || docs.length === 0) {
              break
            }
            await client.bulk(
              DOMAIN_CONTACT,
              docs.map((doc) => ({
                filter: { _id: doc._id },
                update: { $set: { description: makeCollaborativeDoc(doc._id, 'description') } }
              }))
            )
          }
          const it2 = await client.traverse(DOMAIN_CONTACT, { [lead.mixin.Customer + '.customerDescription']: null })
          while (true) {
            const docs = await it2.next(50)
            if (docs == null || docs.length === 0) {
              break
            }
            await client.bulk(
              DOMAIN_CONTACT,
              docs.map((doc) => ({
                filter: { _id: doc._id },
                update: {
                  $set: {
                    [lead.mixin.Customer + '.customerDescription']: makeCollaborativeDoc(
                      docs[0]._id,
                      'customerDescription'
                    )
                  }
                }
              }))
            )
          }
        }
      }
    ])
  },
  async upgrade (state: Map<string, Set<string>>, client: () => Promise<MigrationUpgradeClient>): Promise<void> {
    await tryUpgrade(state, client, leadId, [
      {
        state: 'u-default-funnel',
        func: async (client) => {
          const ops = new TxOperations(client, core.account.System)
          await createDefaults(ops)
        }
      }
    ])
  }
}
